from __future__ import annotations

import typing

from fntypes.co import Option, Result, Some, Variative

from telegrinder.api import ABCAPI, APIError
from telegrinder.model import get_params
from telegrinder.msgspec_utils import Nothing
from telegrinder.types import (
    ChatAction,
    DiceEmoji,
    ForceReply,
    InlineKeyboardMarkup,
    InputFile,
    InputMedia,
    InputPollOption,
    LabeledPrice,
    LinkPreviewOptions,
    Message,
    MessageEntity,
    MessageId,
    ReactionEmoji,
    ReactionType,
    ReplyKeyboardMarkup,
    ReplyKeyboardRemove,
    ReplyParameters,
    User,
)

from .base import BaseCute, compose_method_params, shortcut
from .utils import (
    compose_link_preview_options,
    compose_reactions,
    compose_reply_params,
    input_media,
)

if typing.TYPE_CHECKING:
    from datetime import datetime

    from .callback_query import CallbackQueryCute

MediaType: typing.TypeAlias = typing.Literal[
    "animation",
    "audio",
    "document",
    "photo",
    "video",
]
ReplyMarkup: typing.TypeAlias = typing.Union[
    InlineKeyboardMarkup,
    ReplyKeyboardMarkup,
    ReplyKeyboardRemove,
    ForceReply,
]


async def execute_method_answer(
    message: MessageCute,
    method_name: str,
    params: dict[str, typing.Any],
) -> Result[typing.Any, APIError]:
    params = compose_method_params(
        params=params,
        update=message,
        default_params={"chat_id", "message_thread_id"},
        validators={"message_thread_id": lambda x: x.is_topic_message.unwrap_or(False)},
    )
    reply_parameters = params.get("reply_parameters")
    link_preview_options = params.get("link_preview_options")

    if reply_parameters is not None and isinstance(reply_parameters, dict):
        reply_parameters.setdefault("message_id", params.get("message_id", message.message_id))
        reply_parameters.setdefault("chat_id", params.get("chat_id"))
        params["reply_parameters"] = compose_reply_params(**reply_parameters)

    if link_preview_options is not None and isinstance(link_preview_options, dict):
        params["link_preview_options"] = compose_link_preview_options(**link_preview_options)

    result = await getattr(message.ctx_api, method_name)(**params)
    return result.map(
        lambda x: (
            x
            if isinstance(x, bool)
            else (
                MessageCute.from_update(x, bound_api=message.api)
                if not isinstance(x, list)
                else [MessageCute.from_update(m, bound_api=message.api) for m in x]
            )
        )
    )


async def execute_method_reply(
    message: MessageCute,
    method_name: str,
    params: dict[str, typing.Any],
) -> Result[typing.Any, APIError]:
    params.setdefault("reply_parameters", {})
    return await execute_method_answer(message, method_name, params)


async def execute_method_edit(
    update: MessageCute | CallbackQueryCute,
    method_name: str,
    params: dict[str, typing.Any],
) -> Result[typing.Any, APIError]:
    params = compose_method_params(
        params=params,
        update=update,
        default_params={
            "chat_id",
            "message_id",
            "message_thread_id",
            "inline_message_id",
        },
        validators={
            "inline_message_id": lambda x: not x.message_id,
            "message_thread_id": lambda x: (
                x.is_topic_message.unwrap_or(False)
                if isinstance(x, MessageCute)
                else bool(x.message) and getattr(x.message.unwrap().v, "is_topic_message", False)
            ),
        },
    )
    if "inline_message_id" in params:
        params.pop("message_id", None)
        params.pop("chat_id", None)
    result = await getattr(update.ctx_api, method_name)(**params)
    return result.map(
        lambda v: Variative[MessageCute, bool](
            v.only()
            .map(
                lambda x: MessageCute.from_update(x, bound_api=update.api),
            )
            .unwrap_or(typing.cast(bool, v.v))
        )
    )


def get_entity_value(
    entity_value: typing.Literal["user", "url", "custom_emoji_id", "language"],
    entities: Option[list[MessageEntity]] = Nothing,
    caption_entities: Option[list[MessageEntity]] = Nothing,
) -> Option[typing.Any]:
    ents = entities.unwrap_or(caption_entities.unwrap_or_none())
    if not ents:
        return Nothing
    for entity in ents:
        if (obj := getattr(entity, entity_value, Nothing)) is not Nothing:
            return obj if isinstance(obj, Some) else Some(obj)
    return Nothing


class MessageCute(BaseCute[Message], Message, kw_only=True):
    api: ABCAPI

    @property
    def mentioned_user(self) -> Option[User]:
        """Mentioned user without username."""

        return get_entity_value("user", self.entities, self.caption_entities)

    @property
    def url(self) -> Option[str]:
        """Clickable text URL."""

        return get_entity_value("url", self.entities, self.caption_entities)

    @property
    def programming_language(self) -> Option[str]:
        """The programming language of the entity text."""

        return get_entity_value("language", self.entities, self.caption_entities)

    @property
    def custom_emoji_id(self) -> Option[str]:
        """Unique identifier of the custom emoji."""

        return get_entity_value("custom_emoji_id", self.entities, self.caption_entities)

    @shortcut(
        "send_message",
        executor=execute_method_answer,
        custom_params={"link_preview_options", "reply_parameters", "message_thread_id"},
    )
    async def answer(
        self,
        text: str | None = None,
        chat_id: int | str | None = None,
        message_thread_id: int | None = None,
        business_connection_id: str | None = None,
        parse_mode: str | None = None,
        entities: list[MessageEntity] | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        link_preview_options: LinkPreviewOptions | dict[str, typing.Any] | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_message()`, see the [documentation](https://core.telegram.org/bots/api#sendmessage)

        Use this method to send a message with text messages. On success, the sent Message is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param text: Text of the message to be sent, 1-4096 characters after entities parsing. \

        :param parse_mode: Mode for parsing entities in the message text. See formatting options for \
        more details.

        :param entities: A JSON-serialized list of special entities that appear in message text, \
        which can be specified instead of parse_mode.

        :param link_preview_options: Link preview generation options for the message.

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_message",
        executor=execute_method_reply,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def reply(
        self,
        text: str | None = None,
        chat_id: int | str | None = None,
        message_id: int | None = None,
        message_thread_id: int | None = None,
        business_connection_id: str | None = None,
        parse_mode: str | None = None,
        entities: list[MessageEntity] | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        link_preview_options: LinkPreviewOptions | dict[str, typing.Any] | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_message()`, see the [documentation](https://core.telegram.org/bots/api#sendmessage)

        Use this method to send a reply to a message with text messages. On success, the sent Message is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.    
    
        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param text: Text of the message to be sent, 1-4096 characters after entities parsing. \

        :param parse_mode: Mode for parsing entities in the message text. See formatting options for \
        more details.

        :param entities: A JSON-serialized list of special entities that appear in message text, \
        which can be specified instead of parse_mode.

        :param link_preview_options: Link preview generation options for the message.

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut("delete_message", custom_params={"message_thread_id"})
    async def delete(
        self,
        chat_id: int | None = None,
        message_id: int | None = None,
        message_thread_id: int | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.delete_message()`, see the [documentation](https://core.telegram.org/bots/api#deletemessage)

        Use this method to delete a message, including service messages, with the 
        following limitations: - A message can only be deleted if it was sent less 
        than 48 hours ago. - Service messages about a supergroup, channel, or forum 
        topic creation can't be deleted. - A dice message in a private chat can only 
        be deleted if it was sent more than 24 hours ago. - Bots can delete outgoing 
        messages in private chats, groups, and supergroups. - Bots can delete incoming 
        messages in private chats. - Bots granted can_post_messages permissions 
        can delete outgoing messages in channels. - If the bot is an administrator 
        of a group, it can delete any message there. - If the bot has can_delete_messages 
        permission in a supergroup or a channel, it can delete any message there. 
        Returns True on success.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_id: Identifier of the message to delete.

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only."""

        params = compose_method_params(
            params=get_params(locals()),
            update=self,
            default_params={"chat_id", "message_id", "message_thread_id"},
            validators={"message_thread_id": lambda x: x.is_topic_message.unwrap_or(False)},
        )
        return await self.ctx_api.delete_message(**params)

    @shortcut(
        "edit_message_text",
        executor=execute_method_edit,
        custom_params={"link_preview_options", "message_thread_id"},
    )
    async def edit(
        self,
        text: str | None = None,
        chat_id: int | None = None,
        message_id: int | None = None,
        message_thread_id: int | None = None,
        parse_mode: str | None = None,
        entities: list[MessageEntity] | None = None,
        link_preview_options: LinkPreviewOptions | dict[str, typing.Any] | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
        **other: typing.Any,
    ) -> Result[Variative[MessageCute, bool], APIError]:
        """Shortcut `API.edit_message_text()`, see the [documentation](https://core.telegram.org/bots/api#editmessagetext)

        Use this method to edit text and game messages. On success, if the edited 
        message is not an inline message, the edited Message is returned, otherwise 
        True is returned.

        :param chat_id: Required if inline_message_id is not specified. Unique identifier for \
        the target chat or username of the target channel (in the format @channelusername). \

        :param message_id: Required if inline_message_id is not specified. Identifier of the message \
        to edit.

        :param inline_message_id: Required if chat_id and message_id are not specified. Identifier of the \
        inline message.

        :param text: New text of the message, 1-4096 characters after entities parsing.

        :param parse_mode: Mode for parsing entities in the message text. See formatting options for \
        more details.

        :param entities: A JSON-serialized list of special entities that appear in message text, \
        which can be specified instead of parse_mode.

        :param link_preview_options: Link preview generation options for the message.

        :param reply_markup: A JSON-serialized object for an inline keyboard."""

        ...

    @shortcut(
        "copy_message",
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def copy(
        self,
        chat_id: int | str | None = None,
        from_chat_id: int | str | None = None,
        message_id: int | None = None,
        message_thread_id: int | None = None,
        caption: str | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageId, APIError]:
        """Shortcut `API.copy_message()`, see the [documentation](https://core.telegram.org/bots/api#copymessage)

        Use this method to copy messages of any kind. Service messages, giveaway 
        messages, giveaway winners messages, and invoice messages can't be copied. 
        A quiz poll can be copied only if the value of the field correct_option_id 
        is known to the bot. The method is analogous to the method forwardMessage, 
        but the copied message doesn't have a link to the original message. Returns 
        the MessageId of the sent message on success.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param from_chat_id: Unique identifier for the chat where the original message was sent (or channel \
        username in the format @channelusername).

        :param message_id: Message identifier in the chat specified in from_chat_id.

        :param caption: New caption for media, 0-1024 characters after entities parsing. If not \
        specified, the original caption is kept.

        :param parse_mode: Mode for parsing entities in the new caption. See formatting options for \
        more details.

        :param caption_entities: A JSON-serialized list of special entities that appear in the new caption, \
        which can be specified instead of parse_mode.

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user.
        """

        params = compose_method_params(
            params=get_params(locals()),
            update=self,
            default_params={
                "chat_id",
                "message_id",
                ("from_chat_id", "chat_id"),
                "message_thread_id",
            },
            validators={"message_thread_id": lambda x: x.is_topic_message.unwrap_or(False)},
        )
        if isinstance(reply_parameters, dict):
            reply_parameters.setdefault("message_id", params.get("message_id"))
            reply_parameters.setdefault("chat_id", params.get("chat_id"))
            params["reply_parameters"] = compose_reply_params(**reply_parameters)
        return await self.ctx_api.copy_message(**params)

    @shortcut(
        "set_message_reaction",
        custom_params={"message_thread_id"},
    )
    async def react(
        self,
        reaction: (
            str | ReactionEmoji | ReactionType | list[str | ReactionEmoji | ReactionType] | None
        ) = None,
        chat_id: int | str | None = None,
        message_thread_id: int | None = None,
        message_id: int | None = None,
        is_big: bool | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.set_message_reaction()`, see the [documentation](https://core.telegram.org/bots/api#setmessagereaction)

        Use this method to change the chosen reactions on a message. Service messages 
        can't be reacted to. Automatically forwarded messages from a channel to 
        its discussion group have the same available reactions as messages in the 
        channel. Returns True on success.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_id: Identifier of the target message. If the message belongs to a media group, \
        the reaction is set to the first non-deleted message in the group instead. \

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param reaction: New list of reaction types to set on the message. Currently, as non-premium \
        users, bots can set up to one reaction per message. A custom emoji reaction \
        can be used if it is either already present on the message or explicitly allowed \
        by chat administrators.

        :param is_big: Pass True to set the reaction with a big animation.
        """

        params = compose_method_params(
            params=get_params(locals()),
            update=self,
            default_params={"chat_id", "message_id", "message_thread_id"},
            validators={"message_thread_id": lambda x: x.is_topic_message.unwrap_or(False)},
        )
        if reaction:
            params["reaction"] = compose_reactions(
                reaction.unwrap() if isinstance(reaction, Some) else reaction,
            )
        return await self.ctx_api.set_message_reaction(**params)

    @shortcut("forward_message", custom_params={"message_thread_id"})
    async def forward(
        self,
        chat_id: int | str,
        message_id: int | None = None,
        from_chat_id: int | str | None = None,
        message_thread_id: int | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.forward_message()`, see the [documentation](https://core.telegram.org/bots/api#forwardmessage)

        Use this method to forward messages of any kind. Service messages and messages 
        with protected content can't be forwarded. On success, the sent Message 
        is returned.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param from_chat_id: Unique identifier for the chat where the original message was sent (or channel \
        username in the format @channelusername).

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the forwarded message from forwarding and saving. \

        :param message_id: Message identifier in the chat specified in from_chat_id."""

        params = compose_method_params(
            params=get_params(locals()),
            update=self,
            default_params={
                ("from_chat_id", "chat_id"),
                "message_id",
                "message_thread_id",
            },
            validators={"message_thread_id": lambda x: x.is_topic_message.unwrap_or(False)},
        )
        return (await self.ctx_api.forward_message(**params)).map(
            lambda message: MessageCute.from_update(message, bound_api=self.api),
        )

    @shortcut("pin_chat_message", custom_params={"message_thread_id"})
    async def pin(
        self,
        chat_id: int | str | None = None,
        message_id: int | None = None,
        message_thread_id: int | None = None,
        disable_notification: bool | None = None,
        **other: typing.Any,
    ) -> Result[bool, "APIError"]:
        """Shortcut `API.pin_chat_message()`, see the [documentation](https://core.telegram.org/bots/api#pinchatmessage)

        Use this method to add a message to the list of pinned messages in a chat. If 
        the chat is not a private chat, the bot must be an administrator in the chat 
        for this to work and must have the 'can_pin_messages' administrator right 
        in a supergroup or 'can_edit_messages' administrator right in a channel. 
        Returns True on success.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param message_id: Identifier of a message to pin.

        :param disable_notification: Pass True if it is not necessary to send a notification to all chat members \
        about the new pinned message. Notifications are always disabled in channels \
        and private chats.
        """

        params = compose_method_params(
            params=get_params(locals()),
            update=self,
            default_params={"chat_id", "message_id", "message_thread_id"},
            validators={"message_thread_id": lambda x: x.is_topic_message.unwrap_or(False)},
        )
        return await self.ctx_api.pin_chat_message(**params)

    @shortcut("unpin_chat_message", custom_params={"message_thread_id"})
    async def unpin(
        self,
        chat_id: int | str | None = None,
        message_id: int | None = None,
        message_thread_id: int | None = None,
        **other: typing.Any,
    ) -> Result[bool, "APIError"]:
        """Shortcut `API.unpin_chat_message()`, see the [documentation](https://core.telegram.org/bots/api#unpinchatmessage)

        Use this method to remove a message from the list of pinned messages in a chat. 
        If the chat is not a private chat, the bot must be an administrator in the chat 
        for this to work and must have the 'can_pin_messages' administrator right 
        in a supergroup or 'can_edit_messages' administrator right in a channel. 
        Returns True on success.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_id: Identifier of a message to unpin. If not specified, the most recent pinned \
        message (by sending date) will be unpinned.
        """

        params = compose_method_params(
            params=get_params(locals()),
            update=self,
            default_params={"chat_id", "message_id", "message_thread_id"},
            validators={"message_thread_id": lambda x: x.is_topic_message.unwrap_or(False)},
        )
        return await self.ctx_api.pin_chat_message(**params)

    @shortcut(
        "send_audio",
        executor=execute_method_answer,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def answer_audio(
        self,
        audio: InputFile | str,
        chat_id: int | str | None = None,
        message_thread_id: int | None = None,
        business_connection_id: str | None = None,
        caption: str | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        duration: int | None = None,
        performer: str | None = None,
        title: str | None = None,
        thumbnail: InputFile | str | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_audio()`, see the [documentation](https://core.telegram.org/bots/api#sendaudio)

        Use this method to send audio files, if you want Telegram clients to display 
        them in the music player. Your audio must be in the .MP3 or .M4A format. On 
        success, the sent Message is returned. Bots can currently send audio files 
        of up to 50 MB in size, this limit may be changed in the future. For sending 
        voice messages, use the sendVoice method instead.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param audio: Audio file to send. Pass a file_id as String to send an audio file that exists \
        on the Telegram servers (recommended), pass an HTTP URL as a String for Telegram \
        to get an audio file from the Internet, or upload a new one using multipart/form-data. \
        More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \

        :param caption: Audio caption, 0-1024 characters after entities parsing.

        :param parse_mode: Mode for parsing entities in the audio caption. See formatting options \
        for more details.

        :param caption_entities: A JSON-serialized list of special entities that appear in the caption, \
        which can be specified instead of parse_mode.

        :param duration: Duration of the audio in seconds.

        :param performer: Performer.

        :param title: Track name.

        :param thumbnail: Thumbnail of the file sent; can be ignored if thumbnail generation for the \
        file is supported server-side. The thumbnail should be in JPEG format and \
        less than 200 kB in size. A thumbnail's width and height should not exceed \
        320. Ignored if the file is not uploaded using multipart/form-data. Thumbnails \
        can't be reused and can be only uploaded as a new file, so you can pass `attach://<file_attach_name>` \
        if the thumbnail was uploaded using multipart/form-data under <file_attach_name>. \
        More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_animation",
        executor=execute_method_answer,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def answer_animation(
        self,
        animation: InputFile | str,
        chat_id: int | str | None = None,
        message_thread_id: int | None = None,
        business_connection_id: str | None = None,
        caption: str | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        duration: int | None = None,
        width: int | None = None,
        height: int | None = None,
        thumbnail: InputFile | str | None = None,
        has_spoiler: bool | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_animation()`, see the [documentation](https://core.telegram.org/bots/api#sendanimation)

        Use this method to send animation files (GIF or H.264/MPEG-4 AVC video without 
        sound). On success, the sent Message is returned. Bots can currently send 
        animation files of up to 50 MB in size, this limit may be changed in the future.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param animation: Animation to send. Pass a file_id as String to send an animation that exists \
        on the Telegram servers (recommended), pass an HTTP URL as a String for Telegram \
        to get an animation from the Internet, or upload a new animation using multipart/form-data. \
        More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \

        :param duration: Duration of sent animation in seconds.

        :param width: Animation width.

        :param height: Animation height.

        :param thumbnail: Thumbnail of the file sent; can be ignored if thumbnail generation for the \
        file is supported server-side. The thumbnail should be in JPEG format and \
        less than 200 kB in size. A thumbnail's width and height should not exceed \
        320. Ignored if the file is not uploaded using multipart/form-data. Thumbnails \
        can't be reused and can be only uploaded as a new file, so you can pass `attach://<file_attach_name>` \
        if the thumbnail was uploaded using multipart/form-data under <file_attach_name>. \
        More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \

        :param caption: Animation caption (may also be used when resending animation by file_id), \
        0-1024 characters after entities parsing.

        :param parse_mode: Mode for parsing entities in the animation caption. See formatting options \
        for more details.

        :param caption_entities: A JSON-serialized list of special entities that appear in the caption, \
        which can be specified instead of parse_mode.

        :param has_spoiler: Pass True if the animation needs to be covered with a spoiler animation. \

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_document",
        executor=execute_method_answer,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def answer_document(
        self,
        document: InputFile | str,
        chat_id: int | str | None = None,
        message_thread_id: int | None = None,
        business_connection_id: str | None = None,
        caption: str | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        disable_content_type_detection: bool | None = None,
        thumbnail: InputFile | str | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_document()`, see the [documentation](https://core.telegram.org/bots/api#senddocument)

        Use this method to send general files. On success, the sent Message is returned. 
        Bots can currently send files of any type of up to 50 MB in size, this limit 
        may be changed in the future.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param document: File to send. Pass a file_id as String to send a file that exists on the Telegram \
        servers (recommended), pass an HTTP URL as a String for Telegram to get a \
        file from the Internet, or upload a new one using multipart/form-data. \
        More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \

        :param thumbnail: Thumbnail of the file sent; can be ignored if thumbnail generation for the \
        file is supported server-side. The thumbnail should be in JPEG format and \
        less than 200 kB in size. A thumbnail's width and height should not exceed \
        320. Ignored if the file is not uploaded using multipart/form-data. Thumbnails \
        can't be reused and can be only uploaded as a new file, so you can pass `attach://<file_attach_name>` \
        if the thumbnail was uploaded using multipart/form-data under <file_attach_name>. \
        More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \

        :param caption: Document caption (may also be used when resending documents by file_id), \
        0-1024 characters after entities parsing.

        :param parse_mode: Mode for parsing entities in the document caption. See formatting options \
        for more details.

        :param caption_entities: A JSON-serialized list of special entities that appear in the caption, \
        which can be specified instead of parse_mode.

        :param disable_content_type_detection: Disables automatic server-side content type detection for files uploaded \
        using multipart/form-data.

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_photo",
        executor=execute_method_answer,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def answer_photo(
        self,
        photo: InputFile | str,
        chat_id: int | str | None = None,
        message_thread_id: int | None = None,
        business_connection_id: str | None = None,
        caption: str | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        has_spoiler: bool | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_photo()`, see the [documentation](https://core.telegram.org/bots/api#sendphoto)

        Use this method to send photos. On success, the sent Message is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param photo: Photo to send. Pass a file_id as String to send a photo that exists on the Telegram \
        servers (recommended), pass an HTTP URL as a String for Telegram to get a \
        photo from the Internet, or upload a new photo using multipart/form-data. \
        The photo must be at most 10 MB in size. The photo's width and height must not \
        exceed 10000 in total. Width and height ratio must be at most 20. More information \
        on Sending Files: https://core.telegram.org/bots/api#sending-files. \

        :param caption: Photo caption (may also be used when resending photos by file_id), 0-1024 \
        characters after entities parsing.

        :param parse_mode: Mode for parsing entities in the photo caption. See formatting options \
        for more details.

        :param caption_entities: A JSON-serialized list of special entities that appear in the caption, \
        which can be specified instead of parse_mode.

        :param has_spoiler: Pass True if the photo needs to be covered with a spoiler animation.

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_sticker",
        executor=execute_method_answer,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def answer_sticker(
        self,
        sticker: InputFile | str,
        chat_id: int | str | None = None,
        emoji: str | None = None,
        message_thread_id: int | None = None,
        business_connection_id: str | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_sticker()`, see the [documentation](https://core.telegram.org/bots/api#sendsticker)

        Use this method to send static .WEBP, animated .TGS, or video .WEBM stickers. 
        On success, the sent Message is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param sticker: Sticker to send. Pass a file_id as String to send a file that exists on the \
        Telegram servers (recommended), pass an HTTP URL as a String for Telegram \
        to get a .WEBP sticker from the Internet, or upload a new .WEBP or .TGS sticker \
        using multipart/form-data. More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \
        Video stickers can only be sent by a file_id. Animated stickers can't be \
        sent via an HTTP URL.

        :param emoji: Emoji associated with the sticker; only for just uploaded stickers.

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_video",
        executor=execute_method_answer,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def answer_video(
        self,
        video: InputFile | str,
        chat_id: int | str | None = None,
        message_thread_id: int | None = None,
        business_connection_id: str | None = None,
        caption: str | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        duration: int | None = None,
        width: int | None = None,
        height: int | None = None,
        thumbnail: InputFile | str | None = None,
        has_spoiler: bool | None = None,
        supports_streaming: bool | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_video()`, see the [documentation](https://core.telegram.org/bots/api#sendvideo)

        Use this method to send video files, Telegram clients support MPEG4 videos 
        (other formats may be sent as Document). On success, the sent Message is 
        returned. Bots can currently send video files of up to 50 MB in size, this 
        limit may be changed in the future.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param video: Video to send. Pass a file_id as String to send a video that exists on the Telegram \
        servers (recommended), pass an HTTP URL as a String for Telegram to get a \
        video from the Internet, or upload a new video using multipart/form-data. \
        More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \

        :param duration: Duration of sent video in seconds.

        :param width: Video width.

        :param height: Video height.

        :param thumbnail: Thumbnail of the file sent; can be ignored if thumbnail generation for the \
        file is supported server-side. The thumbnail should be in JPEG format and \
        less than 200 kB in size. A thumbnail's width and height should not exceed \
        320. Ignored if the file is not uploaded using multipart/form-data. Thumbnails \
        can't be reused and can be only uploaded as a new file, so you can pass `attach://<file_attach_name>` \
        if the thumbnail was uploaded using multipart/form-data under <file_attach_name>. \
        More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \

        :param caption: Video caption (may also be used when resending videos by file_id), 0-1024 \
        characters after entities parsing.

        :param parse_mode: Mode for parsing entities in the video caption. See formatting options \
        for more details.

        :param caption_entities: A JSON-serialized list of special entities that appear in the caption, \
        which can be specified instead of parse_mode.

        :param has_spoiler: Pass True if the video needs to be covered with a spoiler animation.

        :param supports_streaming: Pass True if the uploaded video is suitable for streaming.

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_video_note",
        executor=execute_method_answer,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def answer_video_note(
        self,
        video_note: InputFile | str,
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        duration: int | None = None,
        length: int | None = None,
        message_thread_id: int | None = None,
        thumbnail: InputFile | str | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_video_note()`, see the [documentation](https://core.telegram.org/bots/api#sendvideonote)

        As of v.4.0, Telegram clients support rounded square MPEG4 videos of up 
        to 1 minute long. Use this method to send video messages. On success, the 
        sent Message is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param video_note: Video note to send. Pass a file_id as String to send a video note that exists \
        on the Telegram servers (recommended) or upload a new video using multipart/form-data. \
        More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \
        Sending video notes by a URL is currently unsupported.

        :param duration: Duration of sent video in seconds.

        :param length: Video width and height, i.e. diameter of the video message.

        :param thumbnail: Thumbnail of the file sent; can be ignored if thumbnail generation for the \
        file is supported server-side. The thumbnail should be in JPEG format and \
        less than 200 kB in size. A thumbnail's width and height should not exceed \
        320. Ignored if the file is not uploaded using multipart/form-data. Thumbnails \
        can't be reused and can be only uploaded as a new file, so you can pass `attach://<file_attach_name>` \
        if the thumbnail was uploaded using multipart/form-data under <file_attach_name>. \
        More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_voice",
        executor=execute_method_answer,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def answer_voice(
        self,
        voice: InputFile | str,
        chat_id: int | str | None = None,
        message_thread_id: int | None = None,
        business_connection_id: str | None = None,
        caption: str | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        duration: int | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_voice()`, see the [documentation](https://core.telegram.org/bots/api#sendvoice)

        Use this method to send audio files, if you want Telegram clients to display 
        the file as a playable voice message. For this to work, your audio must be 
        in an .OGG file encoded with OPUS (other formats may be sent as Audio or Document). 
        On success, the sent Message is returned. Bots can currently send voice 
        messages of up to 50 MB in size, this limit may be changed in the future.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param voice: Audio file to send. Pass a file_id as String to send a file that exists on the \
        Telegram servers (recommended), pass an HTTP URL as a String for Telegram \
        to get a file from the Internet, or upload a new one using multipart/form-data. \
        More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \

        :param caption: Voice message caption, 0-1024 characters after entities parsing.

        :param parse_mode: Mode for parsing entities in the voice message caption. See formatting \
        options for more details.

        :param caption_entities: A JSON-serialized list of special entities that appear in the caption, \
        which can be specified instead of parse_mode.

        :param duration: Duration of the voice message in seconds.

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_poll",
        executor=execute_method_answer,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def answer_poll(
        self,
        question: str,
        options: list[InputPollOption],
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        question_parse_mode: str | None = None,
        question_entities: list[MessageEntity] | None = None,
        is_anonymous: bool | None = None,
        type: typing.Literal["quiz", "regular"] | None = None,
        allows_multiple_answers: bool | None = None,
        correct_option_id: int | None = None,
        explanation: str | None = None,
        explanation_parse_mode: str | None = None,
        explanation_entities: list[MessageEntity] | None = None,
        open_period: int | None = None,
        close_date: datetime | int | None = None,
        is_closed: bool | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_poll()`, see the [documentation](https://core.telegram.org/bots/api#sendpoll)

        Use this method to send a native poll. On success, the sent Message is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param question_parse_mode: Mode for parsing entities in the question. See formatting options for more \
        details. Currently, only custom emoji entities are allowed.

        :param question_entities: A JSON-serialized list of special entities that appear in the poll question. \
        It can be specified instead of question_parse_mode.

        :param question: Poll question, 1-300 characters.

        :param options: A JSON-serialized list of 2-10 answer options.

        :param is_anonymous: True, if the poll needs to be anonymous, defaults to True.

        :param type: Poll type, `quiz` or `regular`, defaults to `regular`.

        :param allows_multiple_answers: True, if the poll allows multiple answers, ignored for polls in quiz mode, \
        defaults to False.

        :param correct_option_id: 0-based identifier of the correct answer option, required for polls in \
        quiz mode.

        :param explanation: Text that is shown when a user chooses an incorrect answer or taps on the lamp \
        icon in a quiz-style poll, 0-200 characters with at most 2 line feeds after \
        entities parsing.

        :param explanation_parse_mode: Mode for parsing entities in the explanation. See formatting options for \
        more details.

        :param explanation_entities: A JSON-serialized list of special entities that appear in the poll explanation, \
        which can be specified instead of parse_mode.

        :param open_period: Amount of time in seconds the poll will be active after creation, 5-600. \
        Can't be used together with close_date.

        :param close_date: Point in time (Unix timestamp) when the poll will be automatically closed. \
        Must be at least 5 and no more than 600 seconds in the future. Can't be used \
        together with open_period.

        :param is_closed: Pass True if the poll needs to be immediately closed. This can be useful for \
        poll preview.

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_venue",
        executor=execute_method_answer,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def answer_venue(
        self,
        latitude: float,
        longitude: float,
        title: str,
        address: str,
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        foursquare_id: str | None = None,
        foursquare_type: str | None = None,
        google_place_id: str | None = None,
        google_place_type: str | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_venue()`, see the [documentation](https://core.telegram.org/bots/api#sendvenue)

        Use this method to send information about a venue. On success, the sent Message 
        is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param latitude: Latitude of the venue.

        :param longitude: Longitude of the venue.

        :param title: Name of the venue.

        :param address: Address of the venue.

        :param foursquare_id: Foursquare identifier of the venue.

        :param foursquare_type: Foursquare type of the venue, if known. (For example, `arts_entertainment/default`, \
        `arts_entertainment/aquarium` or `food/icecream`.).

        :param google_place_id: Google Places identifier of the venue.

        :param google_place_type: Google Places type of the venue. (See supported types.).

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_dice",
        executor=execute_method_answer,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def answer_dice(
        self,
        emoji: DiceEmoji | None = None,
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_dice()`, see the [documentation](https://core.telegram.org/bots/api#senddice)

        Use this method to send an animated emoji that will display a random value. 
        On success, the sent Message is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param emoji: Emoji on which the dice throw animation is based. Currently, must be one \
        of `🎲`, `🎯`, `🏀`, `⚽`, `🎳`, or `🎰`. Dice can have values 1-6 for `🎲`, `🎯` and \
        `🎳`, values 1-5 for `🏀` and `⚽`, and values 1-64 for `🎰`. Defaults to `🎲`. \

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_game",
        executor=execute_method_answer,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def answer_game(
        self,
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        game_short_name: str | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_game()`, see the [documentation](https://core.telegram.org/bots/api#sendgame)

        Use this method to send a game. On success, the sent Message is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat.

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param game_short_name: Short name of the game, serves as the unique identifier for the game. Set \
        up your games via @BotFather.

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: A JSON-serialized object for an inline keyboard. If empty, one 'Play game_title' \
        button will be shown. If not empty, the first button must launch the game."""

        ...

    @shortcut(
        "send_invoice",
        executor=execute_method_answer,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def answer_invoice(
        self,
        title: str,
        description: str,
        payload: str,
        provider_token: str,
        currency: str,
        prices: list[LabeledPrice],
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        max_tip_amount: int | None = None,
        suggested_tip_amounts: list[int] | None = None,
        start_parameter: str | None = None,
        provider_data: str | None = None,
        photo_url: str | None = None,
        photo_size: int | None = None,
        photo_width: int | None = None,
        photo_height: int | None = None,
        need_name: bool | None = None,
        need_phone_number: bool | None = None,
        need_email: bool | None = None,
        need_shipping_address: bool | None = None,
        send_phone_number_to_provider: bool | None = None,
        send_email_to_provider: bool | None = None,
        is_flexible: bool | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_invoice()`, see the [documentation](https://core.telegram.org/bots/api#sendinvoice)

        Use this method to send invoices. On success, the sent Message is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param title: Product name, 1-32 characters.

        :param description: Product description, 1-255 characters.

        :param payload: Bot-defined invoice payload, 1-128 bytes. This will not be displayed to \
        the user, use for your internal processes.

        :param provider_token: Payment provider token, obtained via @BotFather.

        :param currency: Three-letter ISO 4217 currency code, see more on currencies.

        :param prices: Price breakdown, a JSON-serialized list of components (e.g. product price, \
        tax, discount, delivery cost, delivery tax, bonus, etc.).

        :param max_tip_amount: The maximum accepted amount for tips in the smallest units of the currency \
        (integer, not float/double). For example, for a maximum tip of US$ 1.45 \
        pass max_tip_amount = 145. See the exp parameter in currencies.json, it \
        shows the number of digits past the decimal point for each currency (2 for \
        the majority of currencies). Defaults to 0.

        :param suggested_tip_amounts: A JSON-serialized array of suggested amounts of tips in the smallest units \
        of the currency (integer, not float/double). At most 4 suggested tip amounts \
        can be specified. The suggested tip amounts must be positive, passed in \
        a strictly increased order and must not exceed max_tip_amount.

        :param start_parameter: Unique deep-linking parameter. If left empty, forwarded copies of the \
        sent message will have a Pay button, allowing multiple users to pay directly \
        from the forwarded message, using the same invoice. If non-empty, forwarded \
        copies of the sent message will have a URL button with a deep link to the bot \
        (instead of a Pay button), with the value used as the start parameter.

        :param provider_data: JSON-serialized data about the invoice, which will be shared with the payment \
        provider. A detailed description of required fields should be provided \
        by the payment provider.

        :param photo_url: URL of the product photo for the invoice. Can be a photo of the goods or a marketing \
        image for a service. People like it better when they see what they are paying \
        for.

        :param photo_size: Photo size in bytes.

        :param photo_width: Photo width.

        :param photo_height: Photo height.

        :param need_name: Pass True if you require the user's full name to complete the order.

        :param need_phone_number: Pass True if you require the user's phone number to complete the order.

        :param need_email: Pass True if you require the user's email address to complete the order. \

        :param need_shipping_address: Pass True if you require the user's shipping address to complete the order. \

        :param send_phone_number_to_provider: Pass True if the user's phone number should be sent to provider.

        :param send_email_to_provider: Pass True if the user's email address should be sent to provider.

        :param is_flexible: Pass True if the final price depends on the shipping method.

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: A JSON-serialized object for an inline keyboard. If empty, one 'Pay total \
        price' button will be shown. If not empty, the first button must be a Pay button."""

        ...

    @shortcut(
        "send_chat_action",
        executor=execute_method_answer,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def answer_chat_action(
        self,
        action: ChatAction,
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.send_chat_action()`, see the [documentation](https://core.telegram.org/bots/api#sendchataction)

        Use this method when you need to tell the user that something is happening 
        on the bot's side. The status is set for 5 seconds or less (when a message arrives 
        from your bot, Telegram clients clear its typing status). Returns True 
        on success. We only recommend using this method when a response from the 
        bot will take a noticeable amount of time to arrive.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread; supergroups only.

        :param action: Type of action to broadcast. Choose one, depending on what the user is about \
        to receive: typing for text messages, upload_photo for photos, record_video \
        or upload_video for videos, record_voice or upload_voice for voice notes, \
        upload_document for general files, choose_sticker for stickers, find_location \
        for location data, record_video_note or upload_video_note for video \
        notes."""

        ...

    @shortcut(
        "send_media_group",
        custom_params={"media", "reply_parameters", "message_thread_id"},
    )
    async def answer_media_group(
        self,
        media: list[InputMedia | tuple[MediaType, InputFile | str]],
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        caption: str | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        **other: typing.Any,
    ) -> Result[list[MessageCute], APIError]:
        """Shortcut `API.send_media_group()`, see the [documentation](https://core.telegram.org/bots/api#sendmediagroup)

        Use this method to send a group of photos, videos, documents or audios as 
        an album. Documents and audio files can be only grouped in an album with messages 
        of the same type. On success, an array of Messages that were sent is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param media: A JSON-serialized array describing messages to be sent, must include 2-10 \
        items.

        :param caption: Audio caption, 0-1024 characters after entities parsing.

        :param parse_mode: Mode for parsing entities in the audio caption. See formatting options \
        for more details.

        :param caption_entities: A JSON-serialized list of special entities that appear in the caption, \
        which can be specified instead of parse_mode.

        :param disable_notification: Sends messages silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent messages from forwarding and saving. \

        :param reply_parameters: Description of the message to reply to."""

        params = get_params(locals())

        for i, m in enumerate(media[:]):
            if isinstance(m, tuple):
                media.insert(
                    i,
                    input_media(  # type: ignore
                        *media.pop(i),  # type: ignore
                        caption=caption,
                        caption_entities=caption_entities,
                        parse_mode=parse_mode,
                    ),
                )

        return await execute_method_answer(self, "send_media_group", params)

    @shortcut(
        "send_location",
        executor=execute_method_answer,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def answer_location(
        self,
        latitude: float,
        longitude: float,
        chat_id: int | str | None = None,
        message_thread_id: int | None = None,
        business_connection_id: str | None = None,
        horizontal_accuracy: float | None = None,
        heading: int | None = None,
        live_period: int | None = None,
        proximity_alert_radius: int | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_location()`, see the [documentation](https://core.telegram.org/bots/api#sendlocation)

        Use this method to send point on the map. On success, the sent Message is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param latitude: Latitude of the location.

        :param longitude: Longitude of the location.

        :param horizontal_accuracy: The radius of uncertainty for the location, measured in meters; 0-1500. \

        :param live_period: Period in seconds for which the location will be updated (see Live Locations, \
        should be between 60 and 86400.

        :param heading: For live locations, a direction in which the user is moving, in degrees. \
        Must be between 1 and 360 if specified.

        :param proximity_alert_radius: For live locations, a maximum distance for proximity alerts about approaching \
        another chat member, in meters. Must be between 1 and 100000 if specified. \

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_contact",
        executor=execute_method_answer,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def answer_contact(
        self,
        phone_number: str,
        first_name: str,
        last_name: str | None = None,
        vcard: str | None = None,
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_contact()`, see the [documentation](https://core.telegram.org/bots/api#sendcontact)

        Use this method to send phone contacts. On success, the sent Message is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param phone_number: Contact's phone number.

        :param first_name: Contact's first name.

        :param last_name: Contact's last name.

        :param vcard: Additional data about the contact in the form of a vCard, 0-2048 bytes.

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_audio",
        executor=execute_method_reply,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def reply_audio(
        self,
        audio: InputFile | str,
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        caption: str | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        duration: int | None = None,
        performer: str | None = None,
        title: str | None = None,
        thumbnail: InputFile | str | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_audio()`, see the [documentation](https://core.telegram.org/bots/api#sendaudio)

        Use this method to send a reply to a message with audio files, if you want Telegram clients to display 
        them in the music player. Your audio must be in the .MP3 or .M4A format. On 
        success, the sent Message is returned. Bots can currently send audio files 
        of up to 50 MB in size, this limit may be changed in the future. For sending 
        voice messages, use the sendVoice method instead.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param audio: Audio file to send. Pass a file_id as String to send an audio file that exists \
        on the Telegram servers (recommended), pass an HTTP URL as a String for Telegram \
        to get an audio file from the Internet, or upload a new one using multipart/form-data. \
        More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \

        :param caption: Audio caption, 0-1024 characters after entities parsing.

        :param parse_mode: Mode for parsing entities in the audio caption. See formatting options \
        for more details.

        :param caption_entities: A JSON-serialized list of special entities that appear in the caption, \
        which can be specified instead of parse_mode.

        :param duration: Duration of the audio in seconds.

        :param performer: Performer.

        :param title: Track name.

        :param thumbnail: Thumbnail of the file sent; can be ignored if thumbnail generation for the \
        file is supported server-side. The thumbnail should be in JPEG format and \
        less than 200 kB in size. A thumbnail's width and height should not exceed \
        320. Ignored if the file is not uploaded using multipart/form-data. Thumbnails \
        can't be reused and can be only uploaded as a new file, so you can pass `attach://<file_attach_name>` \
        if the thumbnail was uploaded using multipart/form-data under <file_attach_name>. \
        More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_animation",
        executor=execute_method_reply,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def reply_animation(
        self,
        animation: InputFile | str,
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        caption: str | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        duration: int | None = None,
        width: int | None = None,
        height: int | None = None,
        thumbnail: InputFile | str | None = None,
        has_spoiler: bool | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_animation()`, see the [documentation](https://core.telegram.org/bots/api#sendanimation)

        Use this method to send a reply to a message with animation files (GIF or H.264/MPEG-4 AVC video without 
        sound). On success, the sent Message is returned. Bots can currently send 
        animation files of up to 50 MB in size, this limit may be changed in the future.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param animation: Animation to send. Pass a file_id as String to send an animation that exists \
        on the Telegram servers (recommended), pass an HTTP URL as a String for Telegram \
        to get an animation from the Internet, or upload a new animation using multipart/form-data. \
        More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \

        :param duration: Duration of sent animation in seconds.

        :param width: Animation width.

        :param height: Animation height.

        :param thumbnail: Thumbnail of the file sent; can be ignored if thumbnail generation for the \
        file is supported server-side. The thumbnail should be in JPEG format and \
        less than 200 kB in size. A thumbnail's width and height should not exceed \
        320. Ignored if the file is not uploaded using multipart/form-data. Thumbnails \
        can't be reused and can be only uploaded as a new file, so you can pass `attach://<file_attach_name>` \
        if the thumbnail was uploaded using multipart/form-data under <file_attach_name>. \
        More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \

        :param caption: Animation caption (may also be used when resending animation by file_id), \
        0-1024 characters after entities parsing.

        :param parse_mode: Mode for parsing entities in the animation caption. See formatting options \
        for more details.

        :param caption_entities: A JSON-serialized list of special entities that appear in the caption, \
        which can be specified instead of parse_mode.

        :param has_spoiler: Pass True if the animation needs to be covered with a spoiler animation. \

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_document",
        executor=execute_method_reply,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def reply_document(
        self,
        document: InputFile | str,
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        caption: str | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        disable_content_type_detection: bool | None = None,
        thumbnail: InputFile | str | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_document()`, see the [documentation](https://core.telegram.org/bots/api#senddocument)

        Use this method to send a reply to a message with general files. On success, the sent Message is returned. 
        Bots can currently send files of any type of up to 50 MB in size, this limit 
        may be changed in the future.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param document: File to send. Pass a file_id as String to send a file that exists on the Telegram \
        servers (recommended), pass an HTTP URL as a String for Telegram to get a \
        file from the Internet, or upload a new one using multipart/form-data. \
        More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \

        :param thumbnail: Thumbnail of the file sent; can be ignored if thumbnail generation for the \
        file is supported server-side. The thumbnail should be in JPEG format and \
        less than 200 kB in size. A thumbnail's width and height should not exceed \
        320. Ignored if the file is not uploaded using multipart/form-data. Thumbnails \
        can't be reused and can be only uploaded as a new file, so you can pass `attach://<file_attach_name>` \
        if the thumbnail was uploaded using multipart/form-data under <file_attach_name>. \
        More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \

        :param caption: Document caption (may also be used when resending documents by file_id), \
        0-1024 characters after entities parsing.

        :param parse_mode: Mode for parsing entities in the document caption. See formatting options \
        for more details.

        :param caption_entities: A JSON-serialized list of special entities that appear in the caption, \
        which can be specified instead of parse_mode.

        :param disable_content_type_detection: Disables automatic server-side content type detection for files uploaded \
        using multipart/form-data.

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_photo",
        executor=execute_method_reply,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def reply_photo(
        self,
        photo: InputFile | str,
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        caption: str | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        has_spoiler: bool | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_photo()`, see the [documentation](https://core.telegram.org/bots/api#sendphoto)

        Use this method to send a reply to a message with photos. On success, the sent Message is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param photo: Photo to send. Pass a file_id as String to send a photo that exists on the Telegram \
        servers (recommended), pass an HTTP URL as a String for Telegram to get a \
        photo from the Internet, or upload a new photo using multipart/form-data. \
        The photo must be at most 10 MB in size. The photo's width and height must not \
        exceed 10000 in total. Width and height ratio must be at most 20. More information \
        on Sending Files: https://core.telegram.org/bots/api#sending-files. \

        :param caption: Photo caption (may also be used when resending photos by file_id), 0-1024 \
        characters after entities parsing.

        :param parse_mode: Mode for parsing entities in the photo caption. See formatting options \
        for more details.

        :param caption_entities: A JSON-serialized list of special entities that appear in the caption, \
        which can be specified instead of parse_mode.

        :param has_spoiler: Pass True if the photo needs to be covered with a spoiler animation.

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_sticker",
        executor=execute_method_reply,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def reply_sticker(
        self,
        sticker: InputFile | str,
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        emoji: str | None = None,
        message_thread_id: int | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_sticker()`, see the [documentation](https://core.telegram.org/bots/api#sendsticker)

        Use this method to send a reply to a message with static .WEBP, animated .TGS, or video .WEBM stickers. 
        On success, the sent Message is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param sticker: Sticker to send. Pass a file_id as String to send a file that exists on the \
        Telegram servers (recommended), pass an HTTP URL as a String for Telegram \
        to get a .WEBP sticker from the Internet, or upload a new .WEBP or .TGS sticker \
        using multipart/form-data. More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \
        Video stickers can only be sent by a file_id. Animated stickers can't be \
        sent via an HTTP URL.

        :param emoji: Emoji associated with the sticker; only for just uploaded stickers.

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_video",
        executor=execute_method_reply,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def reply_video(
        self,
        video: InputFile | str,
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        caption: str | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        duration: int | None = None,
        width: int | None = None,
        height: int | None = None,
        thumbnail: InputFile | str | None = None,
        has_spoiler: bool | None = None,
        supports_streaming: bool | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_video()`, see the [documentation](https://core.telegram.org/bots/api#sendvideo)

        Use this method to send a reply to a message with video files, Telegram clients support MPEG4 videos 
        (other formats may be sent as Document). On success, the sent Message is 
        returned. Bots can currently send video files of up to 50 MB in size, this 
        limit may be changed in the future.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param video: Video to send. Pass a file_id as String to send a video that exists on the Telegram \
        servers (recommended), pass an HTTP URL as a String for Telegram to get a \
        video from the Internet, or upload a new video using multipart/form-data. \
        More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \

        :param duration: Duration of sent video in seconds.

        :param width: Video width.

        :param height: Video height.

        :param thumbnail: Thumbnail of the file sent; can be ignored if thumbnail generation for the \
        file is supported server-side. The thumbnail should be in JPEG format and \
        less than 200 kB in size. A thumbnail's width and height should not exceed \
        320. Ignored if the file is not uploaded using multipart/form-data. Thumbnails \
        can't be reused and can be only uploaded as a new file, so you can pass `attach://<file_attach_name>` \
        if the thumbnail was uploaded using multipart/form-data under <file_attach_name>. \
        More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \

        :param caption: Video caption (may also be used when resending videos by file_id), 0-1024 \
        characters after entities parsing.

        :param parse_mode: Mode for parsing entities in the video caption. See formatting options \
        for more details.

        :param caption_entities: A JSON-serialized list of special entities that appear in the caption, \
        which can be specified instead of parse_mode.

        :param has_spoiler: Pass True if the video needs to be covered with a spoiler animation.

        :param supports_streaming: Pass True if the uploaded video is suitable for streaming.

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_video_note",
        executor=execute_method_reply,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def reply_video_note(
        self,
        video_note: InputFile | str,
        chat_id: int | str | None = None,
        duration: int | None = None,
        length: int | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        thumbnail: InputFile | str | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_video_note()`, see the [documentation](https://core.telegram.org/bots/api#sendvideonote)

        As of v.4.0, Telegram clients support rounded square MPEG4 videos of up 
        to 1 minute long. Use this method to send a reply to a message with video messages. On success, the 
        sent Message is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param video_note: Video note to send. Pass a file_id as String to send a video note that exists \
        on the Telegram servers (recommended) or upload a new video using multipart/form-data. \
        More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \
        Sending video notes by a URL is currently unsupported.

        :param duration: Duration of sent video in seconds.

        :param length: Video width and height, i.e. diameter of the video message.

        :param thumbnail: Thumbnail of the file sent; can be ignored if thumbnail generation for the \
        file is supported server-side. The thumbnail should be in JPEG format and \
        less than 200 kB in size. A thumbnail's width and height should not exceed \
        320. Ignored if the file is not uploaded using multipart/form-data. Thumbnails \
        can't be reused and can be only uploaded as a new file, so you can pass `attach://<file_attach_name>` \
        if the thumbnail was uploaded using multipart/form-data under <file_attach_name>. \
        More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_voice",
        executor=execute_method_reply,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def reply_voice(
        self,
        voice: InputFile | str,
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        caption: str | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        duration: int | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_voice()`, see the [documentation](https://core.telegram.org/bots/api#sendvoice)

        Use this method to send a reply to a message with audio files, if you want Telegram clients to display 
        the file as a playable voice message. For this to work, your audio must be 
        in an .OGG file encoded with OPUS (other formats may be sent as Audio or Document). 
        On success, the sent Message is returned. Bots can currently send voice 
        messages of up to 50 MB in size, this limit may be changed in the future.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param voice: Audio file to send. Pass a file_id as String to send a file that exists on the \
        Telegram servers (recommended), pass an HTTP URL as a String for Telegram \
        to get a file from the Internet, or upload a new one using multipart/form-data. \
        More information on Sending Files: https://core.telegram.org/bots/api#sending-files. \

        :param caption: Voice message caption, 0-1024 characters after entities parsing.

        :param parse_mode: Mode for parsing entities in the voice message caption. See formatting \
        options for more details.

        :param caption_entities: A JSON-serialized list of special entities that appear in the caption, \
        which can be specified instead of parse_mode.

        :param duration: Duration of the voice message in seconds.

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_poll",
        executor=execute_method_reply,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def reply_poll(
        self,
        question: str,
        options: list[InputPollOption],
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        question_parse_mode: str | None = None,
        question_entities: list[MessageEntity] | None = None,
        is_anonymous: bool | None = None,
        type: typing.Literal["quiz", "regular"] | None = None,
        allows_multiple_answers: bool | None = None,
        correct_option_id: int | None = None,
        explanation: str | None = None,
        explanation_parse_mode: str | None = None,
        explanation_entities: list[MessageEntity] | None = None,
        open_period: int | None = None,
        close_date: datetime | int | None = None,
        is_closed: bool | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_poll()`, see the [documentation](https://core.telegram.org/bots/api#sendpoll)

        Use this method to send a reply to a message with a native poll. On success, the sent Message is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param question_parse_mode: Mode for parsing entities in the question. See formatting options for more \
        details. Currently, only custom emoji entities are allowed.

        :param question_entities: A JSON-serialized list of special entities that appear in the poll question. \
        It can be specified instead of question_parse_mode.

        :param question: Poll question, 1-300 characters.

        :param options: A JSON-serialized list of 2-10 answer options.

        :param is_anonymous: True, if the poll needs to be anonymous, defaults to True.

        :param type: Poll type, `quiz` or `regular`, defaults to `regular`.

        :param allows_multiple_answers: True, if the poll allows multiple answers, ignored for polls in quiz mode, \
        defaults to False.

        :param correct_option_id: 0-based identifier of the correct answer option, required for polls in \
        quiz mode.

        :param explanation: Text that is shown when a user chooses an incorrect answer or taps on the lamp \
        icon in a quiz-style poll, 0-200 characters with at most 2 line feeds after \
        entities parsing.

        :param explanation_parse_mode: Mode for parsing entities in the explanation. See formatting options for \
        more details.

        :param explanation_entities: A JSON-serialized list of special entities that appear in the poll explanation, \
        which can be specified instead of parse_mode.

        :param open_period: Amount of time in seconds the poll will be active after creation, 5-600. \
        Can't be used together with close_date.

        :param close_date: Point in time (Unix timestamp) when the poll will be automatically closed. \
        Must be at least 5 and no more than 600 seconds in the future. Can't be used \
        together with open_period.

        :param is_closed: Pass True if the poll needs to be immediately closed. This can be useful for \
        poll preview.

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_venue",
        executor=execute_method_reply,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def reply_venue(
        self,
        latitude: float,
        longitude: float,
        title: str,
        address: str,
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        foursquare_id: str | None = None,
        foursquare_type: str | None = None,
        google_place_id: str | None = None,
        google_place_type: str | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_venue()`, see the [documentation](https://core.telegram.org/bots/api#sendvenue)

        Use this method to send a reply to a message with information about a venue. On success, the sent Message 
        is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param latitude: Latitude of the venue.

        :param longitude: Longitude of the venue.

        :param title: Name of the venue.

        :param address: Address of the venue.

        :param foursquare_id: Foursquare identifier of the venue.

        :param foursquare_type: Foursquare type of the venue, if known. (For example, `arts_entertainment/default`, \
        `arts_entertainment/aquarium` or `food/icecream`.).

        :param google_place_id: Google Places identifier of the venue.

        :param google_place_type: Google Places type of the venue. (See supported types.).

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_dice",
        executor=execute_method_reply,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def reply_dice(
        self,
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        emoji: DiceEmoji | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_dice()`, see the [documentation](https://core.telegram.org/bots/api#senddice)

        Use this method to send a reply to a message with an animated emoji that will display a random value. 
        On success, the sent Message is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param emoji: Emoji on which the dice throw animation is based. Currently, must be one \
        of `🎲`, `🎯`, `🏀`, `⚽`, `🎳`, or `🎰`. Dice can have values 1-6 for `🎲`, `🎯` and \
        `🎳`, values 1-5 for `🏀` and `⚽`, and values 1-64 for `🎰`. Defaults to `🎲`. \

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_game",
        executor=execute_method_reply,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def reply_game(
        self,
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        game_short_name: str | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_game()`, see the [documentation](https://core.telegram.org/bots/api#sendgame)

        Use this method to send a reply to a message with a game. On success, the sent Message is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat.

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param game_short_name: Short name of the game, serves as the unique identifier for the game. Set \
        up your games via @BotFather.

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: A JSON-serialized object for an inline keyboard. If empty, one 'Play game_title' \
        button will be shown. If not empty, the first button must launch the game."""

        ...

    @shortcut(
        "send_invoice",
        executor=execute_method_reply,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def reply_invoice(
        self,
        title: str,
        description: str,
        payload: str,
        provider_token: str,
        currency: str,
        prices: list[LabeledPrice],
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        max_tip_amount: int | None = None,
        suggested_tip_amounts: list[int] | None = None,
        start_parameter: str | None = None,
        provider_data: str | None = None,
        photo_url: str | None = None,
        photo_size: int | None = None,
        photo_width: int | None = None,
        photo_height: int | None = None,
        need_name: bool | None = None,
        need_phone_number: bool | None = None,
        need_email: bool | None = None,
        need_shipping_address: bool | None = None,
        send_phone_number_to_provider: bool | None = None,
        send_email_to_provider: bool | None = None,
        is_flexible: bool | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_invoice()`, see the [documentation](https://core.telegram.org/bots/api#sendinvoice)

        Use this method to send a reply to a message with invoices. On success, the sent Message is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param title: Product name, 1-32 characters.

        :param description: Product description, 1-255 characters.

        :param payload: Bot-defined invoice payload, 1-128 bytes. This will not be displayed to \
        the user, use for your internal processes.

        :param provider_token: Payment provider token, obtained via @BotFather.

        :param currency: Three-letter ISO 4217 currency code, see more on currencies.

        :param prices: Price breakdown, a JSON-serialized list of components (e.g. product price, \
        tax, discount, delivery cost, delivery tax, bonus, etc.).

        :param max_tip_amount: The maximum accepted amount for tips in the smallest units of the currency \
        (integer, not float/double). For example, for a maximum tip of US$ 1.45 \
        pass max_tip_amount = 145. See the exp parameter in currencies.json, it \
        shows the number of digits past the decimal point for each currency (2 for \
        the majority of currencies). Defaults to 0.

        :param suggested_tip_amounts: A JSON-serialized array of suggested amounts of tips in the smallest units \
        of the currency (integer, not float/double). At most 4 suggested tip amounts \
        can be specified. The suggested tip amounts must be positive, passed in \
        a strictly increased order and must not exceed max_tip_amount.

        :param start_parameter: Unique deep-linking parameter. If left empty, forwarded copies of the \
        sent message will have a Pay button, allowing multiple users to pay directly \
        from the forwarded message, using the same invoice. If non-empty, forwarded \
        copies of the sent message will have a URL button with a deep link to the bot \
        (instead of a Pay button), with the value used as the start parameter.

        :param provider_data: JSON-serialized data about the invoice, which will be shared with the payment \
        provider. A detailed description of required fields should be provided \
        by the payment provider.

        :param photo_url: URL of the product photo for the invoice. Can be a photo of the goods or a marketing \
        image for a service. People like it better when they see what they are paying \
        for.

        :param photo_size: Photo size in bytes.

        :param photo_width: Photo width.

        :param photo_height: Photo height.

        :param need_name: Pass True if you require the user's full name to complete the order.

        :param need_phone_number: Pass True if you require the user's phone number to complete the order.

        :param need_email: Pass True if you require the user's email address to complete the order. \

        :param need_shipping_address: Pass True if you require the user's shipping address to complete the order. \

        :param send_phone_number_to_provider: Pass True if the user's phone number should be sent to provider.

        :param send_email_to_provider: Pass True if the user's email address should be sent to provider.

        :param is_flexible: Pass True if the final price depends on the shipping method.

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: A JSON-serialized object for an inline keyboard. If empty, one 'Pay total \
        price' button will be shown. If not empty, the first button must be a Pay button."""

        ...

    @shortcut(
        "send_media_group",
        custom_params={
            "media",
            "reply_parameters",
            "message_thread_id",
            "caption",
            "caption_entities",
            "parse_mode",
        },
    )
    async def reply_media_group(
        self,
        media: list[InputMedia | tuple[MediaType, InputFile | str]],
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        caption: str | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        **other: typing.Any,
    ) -> Result[list[MessageCute], APIError]:
        """Shortcut `API.send_media_group()`, see the [documentation](https://core.telegram.org/bots/api#sendmediagroup)

        Use this method to send a reply to a message with a group of photos, videos, documents or audios as 
        an album. Documents and audio files can be only grouped in an album with messages 
        of the same type. On success, an array of Messages that were sent is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param media: A JSON-serialized array describing messages to be sent, must include 2-10 \
        items.

        :param caption: Audio caption, 0-1024 characters after entities parsing.

        :param parse_mode: Mode for parsing entities in the audio caption. See formatting options \
        for more details.

        :param caption_entities: A JSON-serialized list of special entities that appear in the caption, \
        which can be specified instead of parse_mode.

        :param disable_notification: Sends messages silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent messages from forwarding and saving. \

        :param reply_parameters: Description of the message to reply to."""

        params = get_params(locals())
        params.setdefault("reply_parameters", {})
        return await self.answer_media_group(**params)

    @shortcut(
        "send_location",
        executor=execute_method_reply,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def reply_location(
        self,
        latitude: float,
        longitude: float,
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        horizontal_accuracy: float | None = None,
        heading: int | None = None,
        live_period: int | None = None,
        proximity_alert_radius: int | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_location()`, see the [documentation](https://core.telegram.org/bots/api#sendlocation)

        Use this method to send a reply to a message with point on the map. On success, the sent Message is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param latitude: Latitude of the location.

        :param longitude: Longitude of the location.

        :param horizontal_accuracy: The radius of uncertainty for the location, measured in meters; 0-1500. \

        :param live_period: Period in seconds for which the location will be updated (see Live Locations, \
        should be between 60 and 86400.

        :param heading: For live locations, a direction in which the user is moving, in degrees. \
        Must be between 1 and 360 if specified.

        :param proximity_alert_radius: For live locations, a maximum distance for proximity alerts about approaching \
        another chat member, in meters. Must be between 1 and 100000 if specified. \

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "send_contact",
        executor=execute_method_reply,
        custom_params={"reply_parameters", "message_thread_id"},
    )
    async def reply_contact(
        self,
        phone_number: str,
        first_name: str,
        last_name: str | None = None,
        vcard: str | None = None,
        chat_id: int | str | None = None,
        business_connection_id: str | None = None,
        message_thread_id: int | None = None,
        disable_notification: bool | None = None,
        protect_content: bool | None = None,
        reply_parameters: ReplyParameters | dict[str, typing.Any] | None = None,
        reply_markup: ReplyMarkup | None = None,
        **other: typing.Any,
    ) -> Result[MessageCute, APIError]:
        """Shortcut `API.send_contact()`, see the [documentation](https://core.telegram.org/bots/api#sendcontact)

        Use this method to send a reply to a message with phone contacts. On success, the sent Message is returned.

        :param business_connection_id: Unique identifier of the business connection on behalf of which the message \
        will be sent.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param phone_number: Contact's phone number.

        :param first_name: Contact's first name.

        :param last_name: Contact's last name.

        :param vcard: Additional data about the contact in the form of a vCard, 0-2048 bytes.

        :param disable_notification: Sends the message silently. Users will receive a notification with no sound. \

        :param protect_content: Protects the contents of the sent message from forwarding and saving.

        :param reply_parameters: Description of the message to reply to.

        :param reply_markup: Additional interface options. A JSON-serialized object for an inline \
        keyboard, custom reply keyboard, instructions to remove reply keyboard \
        or to force a reply from the user."""

        ...

    @shortcut(
        "edit_message_live_location",
        executor=execute_method_edit,
        custom_params={"message_thread_id"},
    )
    async def edit_live_location(
        self,
        latitude: float,
        longitude: float,
        chat_id: int | str | None = None,
        message_id: int | None = None,
        message_thread_id: int | None = None,
        inline_message_id: str | None = None,
        live_period: int | None = None,
        horizontal_accuracy: float | None = None,
        heading: int | None = None,
        proximity_alert_radius: int | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
        **other: typing.Any,
    ) -> Result[Variative[MessageCute, bool], APIError]:
        """Shortcut `API.edit_message_live_location()`, see the [documentation](https://core.telegram.org/bots/api#editmessagelivelocation)

        Use this method to edit live location messages. A location can be edited 
        until its live_period expires or editing is explicitly disabled by a call 
        to stopMessageLiveLocation. On success, if the edited message is not an 
        inline message, the edited Message is returned, otherwise True is returned.

        :param chat_id: Required if inline_message_id is not specified. Unique identifier for \
        the target chat or username of the target channel (in the format @channelusername). \

        :param message_id: Required if inline_message_id is not specified. Identifier of the message \
        to edit.

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param live_period: New period in seconds during which the location can be updated, starting \
        from the message send date. If 0x7FFFFFFF is specified, then the location \
        can be updated forever. Otherwise, the new value must not exceed the current \
        live_period by more than a day, and the live location expiration date must \
        remain within the next 90 days. If not specified, then live_period remains \
        unchanged.

        :param inline_message_id: Required if chat_id and message_id are not specified. Identifier of the \
        inline message.

        :param latitude: Latitude of new location.

        :param longitude: Longitude of new location.

        :param horizontal_accuracy: The radius of uncertainty for the location, measured in meters; 0-1500. \

        :param heading: Direction in which the user is moving, in degrees. Must be between 1 and 360 \
        if specified.

        :param proximity_alert_radius: The maximum distance for proximity alerts about approaching another chat \
        member, in meters. Must be between 1 and 100000 if specified.

        :param reply_markup: A JSON-serialized object for a new inline keyboard."""

        ...

    @shortcut(
        "edit_message_caption",
        executor=execute_method_edit,
        custom_params={"message_thread_id"},
    )
    async def edit_caption(
        self,
        caption: str,
        chat_id: int | str | None = None,
        message_id: int | None = None,
        message_thread_id: int | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
        **other: typing.Any,
    ) -> Result[Variative[MessageCute, bool], APIError]:
        """Shortcut `API.edit_message_caption()`, see the [documentation](https://core.telegram.org/bots/api#editmessagecaption)

        Use this method to edit captions of messages. On success, if the edited message 
        is not an inline message, the edited Message is returned, otherwise True 
        is returned.

        :param chat_id: Required if inline_message_id is not specified. Unique identifier for \
        the target chat or username of the target channel (in the format @channelusername). \

        :param message_id: Required if inline_message_id is not specified. Identifier of the message \
        to edit.

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param caption: New caption of the message, 0-1024 characters after entities parsing. \

        :param parse_mode: Mode for parsing entities in the message caption. See formatting options \
        for more details.

        :param caption_entities: A JSON-serialized list of special entities that appear in the caption, \
        which can be specified instead of parse_mode.

        :param reply_markup: A JSON-serialized object for an inline keyboard."""

        ...

    @shortcut(
        "edit_message_media",
        custom_params={
            "media",
            "type",
            "message_thread_id",
            "caption",
            "parse_mode",
            "caption_entities",
        },
    )
    async def edit_media(
        self,
        media: InputFile | InputMedia | str,
        type: MediaType | None = None,
        caption: str | None = None,
        parse_mode: str | None = None,
        caption_entities: list[MessageEntity] | None = None,
        chat_id: int | str | None = None,
        message_id: int | None = None,
        message_thread_id: int | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
        **other: typing.Any,
    ) -> Result[Variative[MessageCute, bool], APIError]:
        """Shortcut `API.edit_message_media()`, see the [documentation](https://core.telegram.org/bots/api#editmessagemedia)

        Use this method to edit animation, audio, document, photo, or video messages. 
        If a message is part of a message album, then it can be edited only to an audio 
        for audio albums, only to a document for document albums and to a photo or 
        a video otherwise. When an inline message is edited, a new file can't be uploaded; 
        use a previously uploaded file via its file_id or specify a URL. On success, 
        if the edited message is not an inline message, the edited Message is returned, 
        otherwise True is returned.

        :param chat_id: Required if inline_message_id is not specified. Unique identifier for \
        the target chat or username of the target channel (in the format @channelusername). \

        :param message_id: Required if inline_message_id is not specified. Identifier of the message \
        to edit.

        :param message_thread_id: Unique identifier for the target message thread (topic) of the forum; for \
        forum supergroups only.

        :param media: A JSON-serialized object for a new media content of the message.

        :param caption: Audio caption, 0-1024 characters after entities parsing.

        :param parse_mode: Mode for parsing entities in the audio caption. See formatting options \
        for more details.

        :param caption_entities: A JSON-serialized list of special entities that appear in the caption, \
        which can be specified instead of parse_mode.
        
        :param type: Required if media is not an `str | InputMedia` object. Type of the media, \
        must be one of `photo`, `video`, `animation`, `audio`, `document`.

        :param reply_markup: A JSON-serialized object for a new inline keyboard."""

        params = get_params(locals())

        if not isinstance(media, InputMedia):
            assert (
                type
            ), "parameter 'type' is required, because 'media' is not an 'InputMedia' object."
            params["media"] = input_media(
                params.pop("type"),
                media,
                caption=caption,
                caption_entities=caption_entities,
                parse_mode=parse_mode,
            )

        return await execute_method_edit(self, "edit_message_media", params)

    @shortcut(
        "edit_message_reply_markup",
        executor=execute_method_edit,
        custom_params={"message_thread_id"},
    )
    async def edit_reply_markup(
        self,
        chat_id: int | str | None = None,
        message_id: int | None = None,
        message_thread_id: int | None = None,
        reply_markup: InlineKeyboardMarkup | None = None,
        **other: typing.Any,
    ) -> Result[Variative[MessageCute, bool], APIError]:
        """Shortcut `API.edit_message_reply_markup()`, see the [documentation](https://core.telegram.org/bots/api#editmessagereplymarkup)

        Use this method to edit only the reply markup of messages. On success, if 
        the edited message is not an inline message, the edited Message is returned, 
        otherwise True is returned.

        :param chat_id: Required if inline_message_id is not specified. Unique identifier for \
        the target chat or username of the target channel (in the format @channelusername). \

        :param message_id: Required if inline_message_id is not specified. Identifier of the message \
        to edit.

        :param inline_message_id: Required if chat_id and message_id are not specified. Identifier of the \
        inline message.

        :param reply_markup: A JSON-serialized object for an inline keyboard."""

        ...


__all__ = ("MessageCute",)



import pathlib
import typing
from abc import ABC, abstractmethod

import msgspec
from envparse import env
from fntypes.result import Result

from telegrinder.api.error import APIError
from telegrinder.client import ABCClient

from .error import InvalidTokenError


class Token(str):
    def __new__(cls, token: str) -> typing.Self:
        if token.count(":") != 1 or not token.split(":")[0].isdigit():
            raise InvalidTokenError("Invalid token, it should look like this '123:ABC'.")
        return super().__new__(cls, token)

    def __repr__(self) -> str:
        return f"<Token: {self.bot_id}:{''.join(self.split(':')[-1])[:6]}...>"

    @classmethod
    def from_env(
        cls,
        var_name: str = "BOT_TOKEN",
        *,
        is_read: bool = False,
        path_to_envfile: str | pathlib.Path | None = None,
    ) -> typing.Self:
        if not is_read:
            env.read_envfile(path_to_envfile)
        return cls(env.str(var_name))

    @property
    def bot_id(self) -> int:
        return int(self.split(":")[0])


class ABCAPI(ABC):
    http: ABCClient

    @abstractmethod
    async def request(
        self,
        method: str,
        data: dict[str, typing.Any] | None = None,
        files: dict[str, tuple[str, bytes]] | None = None,
    ) -> Result[list[typing.Any] | dict[str, typing.Any] | bool, APIError]:
        pass

    @abstractmethod
    async def request_raw(
        self,
        method: str,
        data: dict[str, typing.Any] | None = None,
        files: dict[str, tuple[str, bytes]] | None = None,
    ) -> Result[msgspec.Raw, APIError]:
        pass

    @property
    @abstractmethod
    def request_url(self) -> str:
        pass

    @property
    @abstractmethod
    def id(self) -> int:
        pass


__all__ = ("ABCAPI", "Token")



from __future__ import annotations

import dataclasses
from abc import ABC, abstractmethod

import typing_extensions as typing

T = typing.TypeVar("T", default=typing.Any)


@dataclasses.dataclass(repr=False, frozen=True)
class CtxVar(typing.Generic[T]):
    value: T
    const: bool = dataclasses.field(default=False, kw_only=True)

    def __repr__(self) -> str:
        return "<{}(value={!r})>".format(
            ("Const" if self.const else "") + CtxVar.__name__,
            self.value,
        )


@dataclasses.dataclass(repr=False, frozen=True)
class GlobalCtxVar(typing.Generic[T]):
    name: str
    value: T
    const: bool = dataclasses.field(default=False, kw_only=True)

    def __repr__(self) -> str:
        return "<{}({}={})>".format(
            self.__class__.__name__,
            self.name,
            repr(CtxVar(self.value, const=self.const)),
        )

    @classmethod
    def collect(cls, name: str, ctx_value: T | CtxVariable[T]) -> typing.Self:
        ctx_value = (
            CtxVar(ctx_value) if not isinstance(ctx_value, CtxVar | GlobalCtxVar) else ctx_value
        )
        params = ctx_value.__dict__
        params["name"] = name
        return cls(**params)


class ABCGlobalContext(ABC, typing.Generic[T]):
    @abstractmethod
    def __getattr__(self, __name: str) -> typing.Any:
        pass

    @abstractmethod
    def __setattr__(self, __name: str, __value: T | CtxVariable[T]) -> None:
        pass

    @abstractmethod
    def __delattr__(self, __name: str) -> None:
        pass


CtxVariable = CtxVar[T] | GlobalCtxVar[T]


__all__ = (
    "ABCGlobalContext",
    "CtxVar",
    "CtxVariable",
    "GlobalCtxVar",
)



import abc
import typing

from telegrinder.bot.cute_types import ChatJoinRequestCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.adapter import EventAdapter

from .abc import ABCRule

ChatJoinRequest: typing.TypeAlias = ChatJoinRequestCute


class ChatJoinRequestRule(ABCRule[ChatJoinRequestCute], requires=[]):
    adapter = EventAdapter("chat_join_request", ChatJoinRequest)

    @abc.abstractmethod
    async def check(self, event: ChatJoinRequest, ctx: Context) -> bool:
        pass


class HasInviteLink(ChatJoinRequestRule):
    async def check(self, event: ChatJoinRequest, ctx: Context) -> bool:
        return bool(event.invite_link)


class InviteLinkName(ChatJoinRequestRule, requires=[HasInviteLink()]):
    def __init__(self, name: str, /) -> None:
        self.name = name

    async def check(self, event: ChatJoinRequest, ctx: Context) -> bool:
        return event.invite_link.unwrap().name.unwrap_or_none() == self.name


class InviteLinkByCreator(ChatJoinRequestRule, requires=[HasInviteLink()]):
    def __init__(self, creator_id: int, /) -> None:
        self.creator_id = creator_id

    async def check(self, event: ChatJoinRequest, ctx: Context) -> bool:
        return event.invite_link.unwrap().creator.id == self.creator_id


__all__ = (
    "ChatJoinRequestRule",
    "HasInviteLink",
    "InviteLinkByCreator",
    "InviteLinkName",
)



import typing

from fntypes.result import Result

from telegrinder.api.abc import ABCAPI, APIError
from telegrinder.types.objects import ChatJoinRequest, User

from .base import BaseCute, shortcut
from .chat_member_updated import ChatMemberShortcuts, chat_member_interaction


class ChatJoinRequestCute(
    BaseCute[ChatJoinRequest], ChatJoinRequest, ChatMemberShortcuts, kw_only=True
):
    api: ABCAPI

    @property
    def from_user(self) -> User:
        return self.from_

    @property
    def user_id(self) -> int:
        return self.from_user.id

    @shortcut("approve_chat_join_request", executor=chat_member_interaction)
    async def approve(
        self,
        chat_id: int | str | None = None,
        user_id: int | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.approve_chat_join_request()`, see the [documentation](https://core.telegram.org/bots/api#approvechatjoinrequest)

        Use this method to approve a chat join request. The bot must be an administrator 
        in the chat for this to work and must have the can_invite_users administrator 
        right. Returns True on success.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param user_id: Unique identifier of the target user.
        """

        ...

    @shortcut("decline_chat_join_request", executor=chat_member_interaction)
    async def decline(
        self,
        chat_id: int | str | None = None,
        user_id: int | None = None,
        **other: typing.Any,
    ) -> Result[bool, APIError]:
        """Shortcut `API.decline_chat_join_request()`, see the [documentation](https://core.telegram.org/bots/api#declinechatjoinrequest)

        Use this method to decline a chat join request. The bot must be an administrator 
        in the chat for this to work and must have the can_invite_users administrator 
        right. Returns True on success.

        :param chat_id: Unique identifier for the target chat or username of the target channel \
        (in the format @channelusername).

        :param user_id: Unique identifier of the target user.
        """

        ...



import typing
from abc import ABC, abstractmethod


class ABCClient(ABC):
    @abstractmethod
    def __init__(self, *args: typing.Any, **kwargs: typing.Any):
        pass

    @abstractmethod
    async def request_text(
        self,
        url: str,
        method: str = "GET",
        data: dict[str, typing.Any] | None = None,
        **kwargs: typing.Any,
    ) -> str:
        pass

    @abstractmethod
    async def request_json(
        self,
        url: str,
        method: str = "GET",
        data: dict[str, typing.Any] | None = None,
        **kwargs: typing.Any,
    ) -> dict[str, typing.Any]:
        pass

    @abstractmethod
    async def request_content(
        self,
        url: str,
        method: str = "GET",
        data: dict[str, typing.Any] | None = None,
        **kwargs: typing.Any,
    ) -> bytes:
        pass

    @abstractmethod
    async def request_bytes(
        self,
        url: str,
        method: str = "GET",
        data: dict[str, typing.Any] | None = None,
        **kwargs: typing.Any,
    ) -> bytes:
        pass

    @abstractmethod
    async def close(self) -> None:
        pass

    @classmethod
    @abstractmethod
    def get_form(
        cls,
        data: dict[str, typing.Any],
        files: dict[str, tuple[str, bytes]] | None = None,
    ) -> typing.Any:
        pass

    async def __aenter__(self) -> typing.Self:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException],
        exc_val: typing.Any,
        exc_tb: typing.Any,
    ) -> None:
        await self.close()


__all__ = ("ABCClient",)



from telegrinder.bot.cute_types import UpdateCute

from .base import ScalarNode


class UpdateNode(ScalarNode, UpdateCute):
    @classmethod
    async def compose(cls, update: UpdateCute) -> UpdateCute:
        return update


__all__ = ("UpdateNode",)



from .attachment import Attachment, Audio, Photo, Video
from .base import ComposeError, DataNode, Node, ScalarNode
from .composer import NodeCollection, NodeSession, compose_node
from .container import ContainerNode
from .message import MessageNode
from .rule import RuleContext
from .source import Source
from .text import Text
from .tools import generate
from .update import UpdateNode

__all__ = (
    "Attachment",
    "Audio",
    "ComposeError",
    "ContainerNode",
    "DataNode",
    "MessageNode",
    "Node",
    "NodeCollection",
    "NodeSession",
    "Photo",
    "RuleContext",
    "ScalarNode",
    "Source",
    "Text",
    "UpdateNode",
    "Video",
    "compose_node",
    "generate",
)



import typing

from telegrinder.bot.cute_types import ChatMemberUpdatedCute
from telegrinder.types.enums import UpdateType

from .abc import BaseStateView

ChatMemberUpdateType: typing.TypeAlias = typing.Literal[
    UpdateType.CHAT_MEMBER,
    UpdateType.MY_CHAT_MEMBER,
]


class ChatMemberView(BaseStateView[ChatMemberUpdatedCute]):
    def __init__(self, *, update_type: ChatMemberUpdateType | None = None) -> None:
        self.auto_rules = []
        self.handlers = []
        self.middlewares = []
        self.return_manager = None
        self.update_type = update_type

    def get_state_key(self, event: ChatMemberUpdatedCute) -> int | None:
        return event.chat_id


__all__ = ("ChatMemberView",)



import ssl
import typing

import aiohttp
import certifi
from aiohttp import ClientSession, TCPConnector

from telegrinder.client.abc import ABCClient
from telegrinder.modules import JSONModule, json

if typing.TYPE_CHECKING:
    from aiohttp import ClientResponse


class AiohttpClient(ABCClient):
    def __init__(
        self,
        session: ClientSession | None = None,
        json_processing_module: JSONModule | None = None,
        timeout: aiohttp.ClientTimeout | None = None,
        **session_params: typing.Any,
    ) -> None:
        self.session = session
        self.json_processing_module = json_processing_module or json
        self.session_params = session_params
        self.timeout = timeout or aiohttp.ClientTimeout(total=0)

    def __repr__(self) -> str:
        return "<{}: session={!r}, timeout={}, closed={}>".format(
            self.__class__.__name__,
            self.session,
            self.timeout,
            True if self.session is None else self.session.closed,
        )

    async def request_raw(
        self,
        url: str,
        method: str = "GET",
        data: dict[str, typing.Any] | None = None,
        **kwargs: typing.Any,
    ) -> "ClientResponse":
        if not self.session:
            self.session = ClientSession(
                connector=TCPConnector(ssl=ssl.create_default_context(cafile=certifi.where())),
                json_serialize=self.json_processing_module.dumps,
                **self.session_params,
            )
        async with self.session.request(
            url=url,
            method=method,
            data=data,
            timeout=self.timeout,
            **kwargs,
        ) as response:
            await response.read()
            return response

    async def request_json(
        self,
        url: str,
        method: str = "GET",
        data: dict[str, typing.Any] | None = None,
        **kwargs: typing.Any,
    ) -> dict[str, typing.Any]:
        response = await self.request_raw(url, method, data, **kwargs)
        return await response.json(
            encoding="utf-8",
            loads=self.json_processing_module.loads,
            content_type=None,
        )

    async def request_text(
        self,
        url: str,
        method: str = "GET",
        data: dict[str, typing.Any] | aiohttp.FormData | None = None,
        **kwargs: typing.Any,
    ) -> str:
        response = await self.request_raw(url, method, data, **kwargs)  # type: ignore
        return await response.text(encoding="utf-8")

    async def request_bytes(
        self,
        url: str,
        method: str = "GET",
        data: dict[str, typing.Any] | aiohttp.FormData | None = None,
        **kwargs: typing.Any,
    ) -> bytes:
        response = await self.request_raw(url, method, data, **kwargs)  # type: ignore
        if response._body is None:
            await response.read()
        return response._body

    async def request_content(
        self,
        url: str,
        method: str = "GET",
        data: dict[str, typing.Any] | None = None,
        **kwargs: typing.Any,
    ) -> bytes:
        response = await self.request_raw(url, method, data, **kwargs)
        return response._body

    async def close(self) -> None:
        if self.session and not self.session.closed:
            await self.session.close()

    @classmethod
    def get_form(
        cls,
        data: dict[str, typing.Any],
        files: dict[str, tuple[str, bytes]] | None = None,
    ) -> aiohttp.formdata.FormData:
        files = files or {}
        form = aiohttp.formdata.FormData(quote_fields=False)
        for k, v in data.items():
            form.add_field(k, str(v))

        for n, f in files.items():
            form.add_field(n, f[1], filename=f[0])

        return form

    def __del__(self) -> None:
        if self.session and not self.session.closed:
            if self.session._connector is not None and self.session._connector_owner:
                self.session._connector.close()
            self.session._connector = None


__all__ = ("AiohttpClient",)



import typing

import vbml

from telegrinder.bot.dispatch.context import Context
from telegrinder.tools.global_context import TelegrinderCtx

from .abc import Message
from .text import TextMessageRule

PatternLike: typing.TypeAlias = str | vbml.Pattern
global_ctx = TelegrinderCtx()


def check_string(patterns: list[vbml.Pattern], s: str, ctx: Context) -> bool:
    for pattern in patterns:
        match global_ctx.vbml_patcher.check(pattern, s):
            case None | False:
                continue
            case {**response}:
                ctx |= response
        return True
    return False


class Markup(TextMessageRule):
    def __init__(self, patterns: PatternLike | list[PatternLike], /):
        if not isinstance(patterns, list):
            patterns = [patterns]
        self.patterns = [
            vbml.Pattern(pattern) if isinstance(pattern, str) else pattern for pattern in patterns
        ]

    async def check(self, message: Message, ctx: Context) -> bool:
        return check_string(self.patterns, message.text.unwrap(), ctx)


__all__ = ("Markup", "check_string")



import dataclasses
import inspect
import typing
from functools import wraps

from fntypes.result import Result

from telegrinder.api import ABCAPI, API
from telegrinder.model import Model, get_params

F = typing.TypeVar("F", bound=typing.Callable[..., typing.Any])
CuteT = typing.TypeVar("CuteT", bound="BaseCute")
UpdateT = typing.TypeVar("UpdateT", bound=Model)

Executor: typing.TypeAlias = typing.Callable[
    [CuteT, str, dict[str, typing.Any]],
    typing.Awaitable[Result[typing.Any, typing.Any]],
]

if typing.TYPE_CHECKING:

    class BaseCute(Model, typing.Generic[UpdateT]):
        api: ABCAPI

        @classmethod
        def from_update(cls, update: UpdateT, bound_api: ABCAPI) -> typing.Self: ...

        @property
        def ctx_api(self) -> API: ...

else:

    class BaseCute(typing.Generic[UpdateT]):
        @classmethod
        def from_update(cls, update, bound_api):
            return cls(**update.to_dict(), api=bound_api)

        @property
        def ctx_api(self):
            assert isinstance(self.api, API)
            return self.api

        def to_dict(self, *, exclude_fields=None):
            exclude_fields = exclude_fields or set()
            return super().to_dict(exclude_fields={"api"} | exclude_fields)


def compose_method_params(
    params: dict[str, typing.Any],
    update: CuteT,
    *,
    default_params: set[str | tuple[str, str]] | None = None,
    validators: dict[str, typing.Callable[[CuteT], bool]] | None = None,
) -> dict[str, typing.Any]:
    """Compose method `params` from `update` by `default_params` and `validators`.
    
    :param params: Method params.
    :param update: Update object.
    :param default_params: Default params. \
    type (`str`) - Attribute name to be taken from `update` if param undefined. \
    type (`tuple[str, str]`) - tuple[0] Parameter name to be set in `params`, \
    tuple[1] attribute name to be taken from `update`.
    :param validators: Validators mapping (`str, Callable`), key - `Parameter name` \
    for which the validator will be applied, value - `Validator`, if returned `True` \
    parameter will be set, otherwise will not be set.
    :return: Composed params.
    """

    default_params = default_params or set()
    validators = validators or {}

    for param in default_params:
        param_name = param if isinstance(param, str) else param[0]
        if param_name not in params:
            if param_name in validators and not validators[param_name](update):
                continue
            params[param_name] = getattr(update, param if isinstance(param, str) else param[1])

    return params


# NOTE: implement parser on ast for methods decorated this decorator
# to support updates to the schema Bot API.
def shortcut(
    method_name: str,
    *,
    executor: Executor[CuteT] | None = None,
    custom_params: set[str] | None = None,
):
    def wrapper(func: F) -> F:
        @wraps(func)
        async def inner(
            self: CuteT,
            *args: typing.Any,
            **kwargs: typing.Any,
        ) -> typing.Any:
            if executor is None:
                return await func(self, *args, **kwargs)
            signature_params = {
                k: p for k, p in inspect.signature(func).parameters.items() if k != "self"
            }
            params: dict[str, typing.Any] = {}
            index = 0

            for k, p in signature_params.items():
                if p.kind in (p.POSITIONAL_OR_KEYWORD, p.POSITIONAL_ONLY) and len(args) > index:
                    params[k] = args[index]
                    index += 1
                    continue
                if p.kind in (p.VAR_KEYWORD, p.VAR_POSITIONAL):
                    params[k] = kwargs.copy() if p.kind is p.VAR_KEYWORD else args[index:]
                    continue
                params[k] = kwargs.pop(k, p.default) if p.default is not p.empty else kwargs.pop(k)

            return await executor(self, method_name, get_params(params))

        inner.__shortcut__ = Shortcut(  # type: ignore
            method_name=method_name,
            executor=executor,
            custom_params=custom_params or set(),
        )
        return inner  # type: ignore

    return wrapper


@dataclasses.dataclass
class Shortcut:
    method_name: str
    _: dataclasses.KW_ONLY
    executor: Executor | None = dataclasses.field(default=None)
    custom_params: set[str] = dataclasses.field(default_factory=lambda: set())


__all__ = ("BaseCute", "Shortcut", "compose_method_params", "shortcut")



from fntypes.result import Ok, Result

from telegrinder.api.abc import ABCAPI
from telegrinder.bot.cute_types.update import UpdateCute
from telegrinder.bot.rules.adapter.abc import ABCAdapter
from telegrinder.bot.rules.adapter.errors import AdapterError
from telegrinder.types.objects import Update


class RawUpdateAdapter(ABCAdapter[Update, UpdateCute]):
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: adapt Update -> UpdateCute>"

    async def adapt(
        self,
        api: ABCAPI,
        update: Update,
    ) -> Result[UpdateCute, AdapterError]:
        if not isinstance(update, UpdateCute):
            return Ok(UpdateCute.from_update(update, api))
        return Ok(update)


__all__ = ("RawUpdateAdapter",)



import abc
import inspect
import typing
from contextlib import suppress

import msgspec

from telegrinder.bot.cute_types import CallbackQueryCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.adapter import EventAdapter
from telegrinder.model import decoder
from telegrinder.tools.buttons import DataclassInstance

from .abc import ABCRule
from .markup import Markup, PatternLike, check_string

CallbackQuery: typing.TypeAlias = CallbackQueryCute
Validator: typing.TypeAlias = typing.Callable[[typing.Any], bool | typing.Awaitable[bool]]
MapDict: typing.TypeAlias = dict[str, "typing.Any | type[typing.Any] | Validator | list[MapDict] | MapDict"]
CallbackMap: typing.TypeAlias = list[tuple[str, "typing.Any | type[typing.Any] | Validator | CallbackMap"]]
CallbackMapStrict: typing.TypeAlias = list[tuple[str, "Validator | CallbackMapStrict"]]


class CallbackQueryRule(ABCRule[CallbackQuery], abc.ABC):
    adapter = EventAdapter("callback_query", CallbackQuery)

    @abc.abstractmethod
    async def check(self, event: CallbackQuery, ctx: Context) -> bool:
        pass


class HasData(CallbackQueryRule):
    async def check(self, event: CallbackQuery, ctx: Context) -> bool:
        return bool(event.data or event.data.unwrap())


class CallbackQueryDataRule(CallbackQueryRule, abc.ABC, requires=[HasData()]):
    pass


class CallbackDataMap(CallbackQueryDataRule):
    def __init__(self, mapping: MapDict, /) -> None:
        self.mapping = self.transform_to_callbacks(
            self.transform_to_map(mapping),
        )

    @classmethod
    def transform_to_map(cls, mapping: MapDict) -> CallbackMap:
        """Transforms MapDict to CallbackMap."""

        callback_map = []

        for k, v in mapping.items():
            if isinstance(v, dict):
                v = cls.transform_to_map(v)
            callback_map.append((k, v))

        return callback_map

    @classmethod
    def transform_to_callbacks(cls, callback_map: CallbackMap) -> CallbackMapStrict:
        """Transforms `CallbackMap` to `CallbackMapStrict`."""

        callback_map_result = []

        for key, value in callback_map:
            if isinstance(value, type):
                validator = (lambda tp: lambda v: isinstance(v, tp))(value)
            elif isinstance(value, list):
                validator = cls.transform_to_callbacks(value)
            elif not callable(value):
                validator = (lambda val: lambda v: val == v)(value)
            else:
                validator = value
            callback_map_result.append((key, validator))

        return callback_map_result

    @staticmethod
    async def run_validator(value: typing.Any, validator: Validator) -> bool:
        """Run async or sync validator."""

        with suppress(BaseException):
            result = validator(value)
            if inspect.isawaitable(result):
                result = await result
            return result

        return False

    @classmethod
    async def match(cls, callback_data: dict[str, typing.Any], callback_map: CallbackMapStrict) -> bool:
        """Matches callback_data with callback_map recursively."""

        for key, validator in callback_map:
            if key not in callback_data:
                return False

            if isinstance(validator, list):
                if not (
                    isinstance(callback_data[key], dict)
                    and await cls.match(callback_data[key], validator)
                ):
                    return False

            elif not await cls.run_validator(callback_data[key], validator):
                return False

        return True

    async def check(self, event: CallbackQuery, ctx: Context) -> bool:
        callback_data = event.decode_callback_data().unwrap_or_none()
        if callback_data is None:
            return False
        if await self.match(callback_data, self.mapping):
            ctx.update(callback_data)
            return True
        return False


class CallbackDataEq(CallbackQueryDataRule):
    def __init__(self, value: str, /) -> None:
        self.value = value

    async def check(self, event: CallbackQuery, ctx: Context) -> bool:
        return event.data.unwrap() == self.value


class CallbackDataJsonEq(CallbackQueryDataRule):
    def __init__(self, d: dict[str, typing.Any], /) -> None:
        self.d = d

    async def check(self, event: CallbackQuery, ctx: Context) -> bool:
        return event.decode_callback_data().unwrap_or_none() == self.d


class CallbackDataJsonModel(CallbackQueryDataRule):
    def __init__(self, model: type[msgspec.Struct] | type[DataclassInstance], /) -> None:
        self.model = model

    async def check(self, event: CallbackQuery, ctx: Context) -> bool:
        with suppress(BaseException):
            ctx.data = decoder.decode(event.data.unwrap().encode(), type=self.model)
            return True
        return False


class CallbackDataMarkup(CallbackQueryDataRule):
    def __init__(self, patterns: PatternLike | list[PatternLike], /) -> None:
        self.patterns = Markup(patterns).patterns

    async def check(self, event: CallbackQuery, ctx: Context) -> bool:
        return check_string(self.patterns, event.data.unwrap(), ctx)


__all__ = (
    "CallbackDataEq",
    "CallbackDataJsonEq",
    "CallbackDataJsonModel",
    "CallbackDataMap",
    "CallbackDataMarkup",
    "CallbackQueryDataRule",
    "CallbackQueryRule",
    "HasData",
)



import enum
import typing
from abc import ABC, abstractmethod


class ABCI18n(ABC):
    @abstractmethod
    def get_translator_by_locale(self, locale: str) -> "ABCTranslator":
        pass


class ABCTranslator(ABC):
    def __init__(self, locale: str, **kwargs: typing.Any) -> None:
        self.locale = locale

    @abstractmethod
    def get(self, __key: str, *args: typing.Any, **kwargs: typing.Any) -> str:
        """This translates a key to actual human-readable string"""

    def __call__(self, __key: str, *args: typing.Any, **kwargs: typing.Any) -> str:
        return self.get(__key, *args, **kwargs)


class I18nEnum(enum.Enum):
    I18N = "_"


__all__ = (
    "ABCI18n",
    "ABCTranslator",
    "I18nEnum",
)



from telegrinder.bot.cute_types.update import UpdateCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.types.enums import UpdateType

from .abc import ABCRule, T


class IsUpdate(ABCRule[T]):
    def __init__(self, update_type: UpdateType, /) -> None:
        self.update_type = update_type

    async def check(self, event: UpdateCute, ctx: Context) -> bool:
        return event.update_type.unwrap_or_none() == self.update_type


__all__ = ("IsUpdate",)



from .abc import (
    ABCReturnManager,
    BaseReturnManager,
    Manager,
    register_manager,
)
from .callback_query import CallbackQueryReturnManager
from .inline_query import InlineQueryReturnManager
from .message import MessageReturnManager

__all__ = (
    "ABCReturnManager",
    "BaseReturnManager",
    "CallbackQueryReturnManager",
    "InlineQueryReturnManager",
    "Manager",
    "MessageReturnManager",
    "register_manager",
)



import asyncio
import contextlib
import dataclasses
import typing

from telegrinder.modules import logger

from .abc import ABCLoopWrapper

T = typing.TypeVar("T")
P = typing.ParamSpec("P")
CoroFunc = typing.TypeVar("CoroFunc", bound="CoroutineFunc")

CoroutineTask: typing.TypeAlias = typing.Coroutine[typing.Any, typing.Any, T]
CoroutineFunc: typing.TypeAlias = typing.Callable[P, CoroutineTask[T]]
Task: typing.TypeAlias = typing.Union[CoroutineFunc, CoroutineTask, "DelayedTask"]


def run_tasks(
    tasks: list[CoroutineTask[typing.Any]],
    loop: asyncio.AbstractEventLoop,
) -> None:
    while tasks:
        loop.run_until_complete(tasks.pop(0))


def to_coroutine_task(task: Task) -> CoroutineTask[typing.Any]:
    if asyncio.iscoroutinefunction(task) or isinstance(task, DelayedTask):
        task = task()
    elif not asyncio.iscoroutine(task):
        raise TypeError("Task should be coroutine or coroutine function.")
    return task


@dataclasses.dataclass
class DelayedTask(typing.Generic[CoroFunc]):
    handler: CoroFunc
    seconds: float
    repeat: bool = dataclasses.field(default=False, kw_only=True)
    _cancelled: bool = dataclasses.field(default=False, init=False, repr=False)

    @property
    def is_cancelled(self) -> bool:
        return self._cancelled

    async def __call__(self, *args, **kwargs) -> None:
        while not self.is_cancelled:
            await asyncio.sleep(self.seconds)
            if self.is_cancelled:
                break
            await self.handler(*args, **kwargs)
            if not self.repeat:
                break

    def cancel(self) -> None:
        if not self._cancelled:
            self._cancelled = True


@dataclasses.dataclass(kw_only=True)
class Lifespan:
    startup_tasks: list[CoroutineTask[typing.Any]] = dataclasses.field(default_factory=lambda: [])
    shutdown_tasks: list[CoroutineTask[typing.Any]] = dataclasses.field(default_factory=lambda: [])

    def on_startup(self, task_or_func: Task) -> Task:
        task_or_func = to_coroutine_task(task_or_func)
        self.startup_tasks.append(task_or_func)
        return task_or_func

    def on_shutdown(self, task_or_func: Task) -> Task:
        task_or_func = to_coroutine_task(task_or_func)
        self.shutdown_tasks.append(task_or_func)
        return task_or_func

    def start(self, loop: asyncio.AbstractEventLoop) -> None:
        run_tasks(self.startup_tasks, loop)

    def shutdown(self, loop: asyncio.AbstractEventLoop) -> None:
        run_tasks(self.shutdown_tasks, loop)


class LoopWrapper(ABCLoopWrapper):
    def __init__(
        self,
        *,
        tasks: list[CoroutineTask[typing.Any]] | None = None,
        lifespan: Lifespan | None = None,
    ) -> None:
        self.tasks: list[CoroutineTask[typing.Any]] = tasks or []
        self.lifespan = lifespan or Lifespan()
        self._loop = asyncio.new_event_loop()

    def __repr__(self) -> str:
        return "<{}: loop={!r} with tasks={!r}, lifespan={!r}>".format(
            self.__class__.__name__,
            self._loop,
            self.tasks,
            self.lifespan,
        )

    def run_event_loop(self) -> None:
        if not self.tasks:
            logger.warning("You run loop with 0 tasks!")

        self.lifespan.start(self._loop)
        while self.tasks:
            self._loop.create_task(self.tasks.pop(0))
        tasks = asyncio.all_tasks(self._loop)

        try:
            while tasks:
                tasks_results, _ = self._loop.run_until_complete(
                    asyncio.wait(tasks, return_when=asyncio.FIRST_EXCEPTION),
                )
                for task_result in tasks_results:
                    try:
                        task_result.result()
                    except BaseException as ex:
                        logger.exception(ex)
                tasks = asyncio.all_tasks(self._loop)
        except KeyboardInterrupt:
            print()  # blank print for ^C
            logger.info("Caught KeyboardInterrupt, cancellation...")
            self.complete_tasks(tasks)
        finally:
            self.lifespan.shutdown(self._loop)
            if self._loop.is_running():
                self._loop.close()

    def add_task(self, task: Task) -> None:
        task = to_coroutine_task(task)

        if self._loop and self._loop.is_running():
            self._loop.create_task(task)
        else:
            self.tasks.append(task)

    def complete_tasks(self, tasks: set[asyncio.Task[typing.Any]]) -> None:
        tasks = tasks | asyncio.all_tasks(self._loop)
        task_to_cancel = asyncio.gather(*tasks, return_exceptions=True)
        task_to_cancel.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            self._loop.run_until_complete(task_to_cancel)

    def timer(
        self,
        *,
        days: int = 0,
        hours: int = 0,
        minutes: int = 0,
        seconds: float = 0,
    ):
        seconds += minutes * 60
        seconds += hours * 60 * 60
        seconds += days * 24 * 60 * 60

        def decorator(func: CoroFunc) -> CoroFunc:
            self.add_task(DelayedTask(func, seconds, repeat=False))
            return func

        return decorator

    def interval(
        self,
        *,
        days: int = 0,
        hours: int = 0,
        minutes: int = 0,
        seconds: float = 0,
    ):
        seconds += minutes * 60
        seconds += hours * 60 * 60
        seconds += days * 24 * 60 * 60

        def decorator(func: CoroFunc) -> CoroFunc:
            self.add_task(DelayedTask(func, seconds, repeat=True))
            return func

        return decorator


__all__ = ("DelayedTask", "Lifespan", "LoopWrapper", "to_coroutine_task")
